from contextlib import contextmanager, ExitStack
import html
from lxml import etree
import re
import requests
from unittest.mock import MagicMock, patch, ANY
from urllib.parse import parse_qsl, urlencode, urlparse
from uuid import uuid4
from odoo.addons.saas_trial.tests.common import TestTrialCommon, new_test_user, tagged
from odoo.addons.saas_worker.const import OAUTH_BASE_URL, TIMEOUT, WEBSITE
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.addons.saas_worker.tests.compat import MockEmail
from odoo.addons.saas_worker.util import version_gte
from odoo.tools import mute_logger


@tagged('post_install', '-at_install', 'post_saas_trial_creation')
class TestTrialUsersInvite(TestTrialCommon):
    def _mock_reset_password(self):
        method = "_action_reset_password" if version_gte("saas~16.4") else "action_reset_password"
        ResUsers_klass = self.env.registry["res.users"]
        return patch.object(ResUsers_klass, method, autospec=True, return_value=True)

    def test_users_invite_monodb(self):
        # ensure employee is considered new (never connected before)
        self.env["res.users.log"].sudo().search([('create_uid', '=', self.user_employee.id)]).unlink()

        with ExitStack() as stack:
            stack.enter_context(self._mock_saas_config(from_role="of", to_role="trial"))
            mock_reset_passwd = stack.enter_context(self._mock_reset_password())

            resp = requests.post(
                self.base_url() + "/saas_trial/create_invitation", auth=SaasAuth(),
                json={'params': {
                    'login': self.user_admin.login,
                    'uid': self.user_admin.id,
                    'secret_signature': self.user_admin._get_secret_signature(),
                    'invitee': {
                        'employee2@example.net': 'Employee 2',
                        'employee@example.net': 'Employee 1',
                    }
                }},
            )
            self.assertEqual(resp.status_code, 200)
            result = resp.json().get('result')
            self.assertTrue(result['result'])
            self.assertEqual(len(result['users']), 2)
            self.assertEqual(
                self.env["res.users"].search([('login', '=', 'employee@example.net')]),
                self.user_employee,
                "Inviting an already existing employee (that never connected) should not crash",
            )
            self.assertTrue(
                self.env["res.users"].search_count([('login', '=', 'employee2@example.net')]),
                "Should have created a new user for employee2@example.net",
            )
            self.assertEqual(
                mock_reset_passwd.call_count, 2,
                "We should only have called reset password for both users"
            )

    def test_users_invite_multidb(self):
        ResUsers_klass = self.env.registry["res.users"]
        db_uuid = self.env["ir.config_parameter"].get_param("database.uuid")

        # ensure employee is considered active (already connected before)
        # (for mutli-db requests, this should not crash)
        self.env(user=self.user_employee)["res.users"]._update_last_login()

        # setup test oauth for admin
        self.user_admin.write({'oauth_uid': '-22'})
        if version_gte("16.0"):
            self.user_admin.flush_recordset()
        else:
            self.user_admin.flush()  # for flush so that other request see the change
        oauth_tokens_info = {
            db_uuid: {"oauth_uid": '-22'},
        }

        with ExitStack() as stack:
            stack.enter_context(self._mock_saas_config(from_role="of", to_role="trial"))
            stack.enter_context(patch("odoo.addons.saas_worker.controllers.main.check_oauth2_tokens",
                                      return_value=oauth_tokens_info))
            mock_reset_passwd = stack.enter_context(self._mock_reset_password())

            resp = requests.post(
                self.base_url() + "/saas_worker/multi/create_invitation/json", auth=SaasAuth(),
                json={'params': {
                    'databases': [
                        {'db_uuid': db_uuid, 'oauth_token': "dummy-test-token"},
                    ],
                    'invitees': [
                        {'mail': 'employee2@example.net', 'name': 'Employee 2'},
                        {'mail': 'employee@example.net', 'name': 'Employee 1'},
                    ],
                }},
            )
            self.assertEqual(resp.status_code, 200)
            result = resp.json().get('result')
            self.assertTrue(result[db_uuid]['result'])
            self.assertEqual(len(result[db_uuid]['users']), 2)
            self.assertEqual(
                self.env["res.users"].search([('login', '=', 'employee@example.net')]),
                self.user_employee,
                "Inviting an already existing employee should not crash",
            )
            self.assertTrue(
                self.env["res.users"].search_count([('login', '=', 'employee2@example.net')]),
                "Should have created a new user for employee2@example.net",
            )
            self.assertEqual(
                mock_reset_passwd.call_count, 0,
                "In multi-db mode, we should not send reset password from invidual databases",
            )


@tagged('post_install', '-at_install', 'post_saas_trial_creation')
class TestTrialUsersInviteMails(TestTrialCommon, MockEmail):
    def setUp(self):
        super().setUp()
        if version_gte("17.0"):
            # since odoo/odoo@0a506d63c Odoo manually rollback savepoint
            # after password reset, so once we return the registered mails
            # do not exist anymore.
            #   As we need to validate the content of those outgoing emails, we
            # skip that manual savepoint closing to allow `MockEmail` helpers
            # to find those emails
            ResUsers = self.env.registry["res.users"]
            _orig_ResUsers_action_reset_password = ResUsers._action_reset_password

            def _action_reset_password(self):
                with patch('contextlib.closing', side_effect=lambda x: x, autospec=True):
                    return _orig_ResUsers_action_reset_password(self)

            stack = ExitStack()
            self.addCleanup(stack.close)
            stack.enter_context(
                patch.object(ResUsers, '_action_reset_password',
                             side_effect=_action_reset_password, autospec=True),
            )

    @contextmanager
    def mock_mail_gateway(self, mail_unlink_sent=False):
        with super().mock_mail_gateway(mail_unlink_sent=mail_unlink_sent):
            if getattr(self, 'testing_smtp_session', None):
                # standard TestingSMTPSession does not have `_host` attribute,
                # but it's need to identify if server is internal or external
                # (see ir.mail_server's `_is_smtp_internal()` method)
                self.testing_smtp_session._host = 'mx.external.example.com'
            yield

    @contextmanager
    def mock_requests_to_reset_password(self):
        original_requests_post = requests.post

        def side_effect(url, *args, json=None, **kwargs):
            if url.startswith(self.base_url()):
                return original_requests_post(url, *args, json=json, **kwargs)

            if url == f'{OAUTH_BASE_URL}/openerp_enterprise/reset_passwords':
                return MagicMock()

            self.fail(f'Call to an invalid route {url!r}')

        with patch('requests.post', side_effect=side_effect) as mock_requests_post:
            yield mock_requests_post

    def assertUserExists(self, login, message):
        user = self.env['res.users'].with_context(active_test=False).search([('login', '=', login)])
        self.assertTrue(user, message)

    def assertUserDoesntExist(self, login, message):
        user = self.env['res.users'].with_context(active_test=False).search([('login', '=', login)])
        self.assertFalse(user, message)

    def assertConfirmUserLink(self, url, dbname, signup_email, token):
        parsed_url = urlparse(url)
        parsed_querystring = dict(parse_qsl(parsed_url.query))
        self.assertDictEqual(parsed_querystring, {
            'db': dbname,
            'signup_email': signup_email,
            'token': token,
        })

    def assertXPath(self, xml, selector, check):
        doc = etree.fromstring(xml, etree.HTMLParser())
        results = doc.xpath(selector)
        if not results:
            self.fail(f"{selector!r} not found in:\n{xml}")
        if len(results) > 1:
            self.fail(f"{selector!r} matches more than one element in:\n{xml}")
        element, = results
        if not element.xpath(check):
            subxml = etree.tostring(element, encoding='UTF-8', xml_declaration=False).decode()
            self.fail(f"{check!r} not verified in:\n{subxml}")

    def web_create_users(self, emails):
        ResUsers = self.env['res.users']
        web_create_users = ResUsers.web_create_users
        return web_create_users(emails)

    def test_invitation_from_database_without_odoocom_account(self):
        self.uid = self.ref('base.user_admin')
        self.assertUserDoesntExist('new_user@domain.tld', 'Invited user should not pre-exist the test')

        with ExitStack() as stack:
            stack.enter_context(self.mock_mail_gateway())
            self._existing_odoocom_users.discard('new_user@domain.tld')
            stack.enter_context(patch('odoo.addons.auth_signup.models.res_partner.random_token',
                                      return_value='r4nd0mT0k3n'))

            self.web_create_users(['new_user@domain.tld'])

        # The confirmation link should point to the db instance directly (standard behavior)
        found_mail = self._find_mail_mail_wemail('new_user@domain.tld', 'sent')
        url_params = {'db': self.cr.dbname}
        if version_gte("16.0"):
            # Introduced by odoo/odoo@97658791306349456761ba4c18d504f861e3bb98 in 16.0
            url_params['signup_email'] = 'new_user@domain.tld'
        url_params['token'] = 'r4nd0mT0k3n'
        url = '/web/reset_password?' + urlencode(url_params)
        self.assertXPath(found_mail.body_html, '//a[contains(text(),"Accept invitation")]',
                         f'substring(./@href,string-length(./@href)-{len(url)}+1)={url!r}')

    def test_invitation_from_database_with_odoocom_account(self):
        self.uid = self.ref('base.user_admin')
        ResUsers = self.env['res.users']
        self.assertUserDoesntExist('new_user@domain.tld', 'Invited user should not pre-exist the test')

        with ExitStack() as stack:
            stack.enter_context(self.mock_mail_gateway())
            self._existing_odoocom_users.add('new_user@domain.tld')
            stack.enter_context(patch('odoo.addons.auth_signup.models.res_partner.random_token',
                                      return_value='r4nd0mT0k3n'))
            self.web_create_users(['new_user@domain.tld'])

            new_user = ResUsers.search([('login', '=', 'new_user@domain.tld')])
            signup_url = new_user.signup_url

        # The confirmation link should point to Odoo website, which will redirect to us after setting the oauth_uid
        found_mail = self._find_mail_mail_wemail('new_user@domain.tld', 'sent')
        self.assertNotIn('/web/reset_password?db=' + self.cr.dbname, found_mail.body_html)
        needle = re.escape(WEBSITE) + r'/openerp_enterprise/confirm_user\?db=' + re.escape(self.cr.dbname) + '[^\'"]*'
        self.assertRegex(found_mail.body_html, needle)

        # The query of the link should keep the standard arguments
        mail_url = html.unescape(re.search(needle, found_mail.body_html).group(0))
        self.assertConfirmUserLink(mail_url, self.cr.dbname, 'new_user@domain.tld', 'r4nd0mT0k3n')

        # The link on the backend form view should be the same
        self.assertConfirmUserLink(signup_url, self.cr.dbname, 'new_user@domain.tld', 'r4nd0mT0k3n')

    def test_reset_password_without_odoocom_account(self):
        self._existing_odoocom_users.discard('existing_user@domain.tld')
        self.web_create_users(['existing_user@domain.tld'])
        user = self.env['res.users'].search([('login', '=', 'existing_user@domain.tld')])

        with ExitStack() as stack:
            stack.enter_context(self.mock_mail_gateway())
            stack.enter_context(patch('odoo.addons.auth_signup.models.res_partner.random_token',
                                      return_value='r4nd0mT0k3n'))
            user.action_reset_password()

        # The confirmation link should point to the db instance directly (standard behavior)
        found_mail = self._find_mail_mail_wemail('existing_user@domain.tld', 'sent')
        url = '/web/reset_password?' + urlencode({
            'db': self.cr.dbname,
            'token': 'r4nd0mT0k3n',
        })
        self.assertXPath(found_mail.body_html, '//a[contains(text(),"Change password")]',
                         f'substring(./@href,string-length(./@href)-{len(url)}+1)={url!r}')

    @patch('odoo.addons.saas_trial.models.res_users.SaasAuth')
    def test_reset_password_with_odoocom_account(self, MockSaasAuth):
        self._existing_odoocom_users.add('existing_user@domain.tld')
        with ExitStack() as stack:
            stack.enter_context(patch('odoo.addons.auth_signup.models.res_partner.random_token',
                                      return_value='v3r153kR3t'))
            mock_requests_post = stack.enter_context(self.mock_requests_to_reset_password())

            self.web_create_users(['existing_user@domain.tld'])
            user = self.env['res.users'].search([('login', '=', 'existing_user@domain.tld')])
            user.signup_token = False  # Also test users that got unlinked and have no valid signup token

            # only catch calls and emails sent by reset_password (user creation just above will ping and send an email)
            stack.enter_context(self.mock_mail_gateway())
            mock_requests_post.reset_mock()

            user.action_reset_password()

        # No email should be sent
        self.assertFalse(self._new_mails)

        # Password reset should be transmitted to Odoo.com
        # The redirect link should put the user back on the confirm_user track
        confirm_user_url = f'{WEBSITE}/openerp_enterprise/confirm_user?' + urlencode({
            'db': self.env.cr.dbname,
            'token': 'v3r153kR3t',
            'signup_email': 'existing_user@domain.tld',
        })
        mock_requests_post.assert_called_once_with(f'{OAUTH_BASE_URL}/openerp_enterprise/reset_passwords', json={
            'params': {'passwords_to_reset': [
                {'login': 'existing_user@domain.tld', 'saas_trial_reset': confirm_user_url},
            ]},
        }, auth=MockSaasAuth(), timeout=TIMEOUT)

    @patch('odoo.addons.saas_trial.models.res_users.SaasAuth')
    def test_reset_password_with_odoocom_linked_user(self, MockSaasAuth):
        with ExitStack() as stack:
            mock_requests_post = stack.enter_context(self.mock_requests_to_reset_password())

            self.web_create_users(['linked_user@domain.tld'])
            user = self.env['res.users'].search([('login', '=', 'linked_user@domain.tld')])
            user.odoo_com_uid = '1337'

            # only catch calls and emails sent by reset_password (user creation just above will ping and send an email)
            stack.enter_context(self.mock_mail_gateway())
            stack.enter_context(patch('odoo.addons.auth_signup.models.res_partner.random_token',
                                      return_value='v3r153kR3t'))
            mock_requests_post.reset_mock()

            user.action_reset_password()

        # No email should be sent
        self.assertFalse(self._new_mails)

        # Password reset should be transmitted to Odoo.com
        # The redirect link should connect the user to their database
        db_uuid = self.env['ir.config_parameter'].sudo().get_param('database.uuid', '')
        confirm_user_url = f'/my/databases/connect/{db_uuid}' if db_uuid else '/my/databases'
        mock_requests_post.assert_called_once_with(f'{OAUTH_BASE_URL}/openerp_enterprise/reset_passwords', json={
            'params': {'passwords_to_reset': [
                {'login': 'linked_user@domain.tld', 'saas_trial_reset': confirm_user_url},
            ]},
        }, auth=MockSaasAuth(), timeout=TIMEOUT)


@tagged('post_install', '-at_install', 'post_saas_trial_creation')
class TestSettingOauthUid(TestTrialCommon):
    def setUp(self):
        super().setUp()

        patch_stack = ExitStack()
        self.addCleanup(patch_stack.close)

        # /saas_trial/set_oauth_uid commits the transaction before calling the publisher warranty
        # mock it into a flush instead, so that the test can safely rollback
        self.flush_all = (self.env.flush_all if version_gte("16.0")
                          else self.env['base'].flush)
        patch_stack.enter_context(patch.object(self.env.cr, 'commit', side_effect=self.flush_all))

        self.user_id = new_test_user(self.env, 'user@doma.in')
        self.user_id.partner_id.signup_prepare()

    def call_set_oauth_uid(self, from_role='accounts', to_role='trial', auth=None, login=None, token=None, oauth_uid='1985'):
        if login is None:
            login = self.user_id.login
        if token is None:
            token = self.user_id.signup_token
        if auth is None:
            auth = SaasAuth()

        payload = {
            'jsonrpc': '2.0',
            'method': 'call',
            'id': 0,
            'params': {
                'login': login,
                'token': token,
                'oauth_uid': oauth_uid,
            },
        }
        with ExitStack() as stack:
            stack.enter_context(self._mock_saas_config(from_role=from_role, to_role=to_role))
            stack.enter_context(mute_logger('odoo.http'))  # might log a Forbidden stacktrace

            response = self.opener.post(self.base_url() + '/saas_trial/set_oauth_uid', json=payload, auth=auth)
            response.raise_for_status()
            return response.json()

    def test_set_oauth_uid_wrong_from_role(self):
        self.assertFalse(self.user_id.oauth_uid)
        data = self.call_set_oauth_uid(from_role='trial')
        self.assertIn('error', data)
        self.assertEqual(data['error']['data']['name'], 'werkzeug.exceptions.Forbidden')

    def test_set_oauth_uid_wrong_login(self):
        self.assertFalse(self.user_id.oauth_uid)
        token = self.user_id.signup_token
        data = self.call_set_oauth_uid(login='unexistant')
        self.assertIn('error', data)
        self.assertEqual(data['error']['data']['message'].strip(), f"Signup token {token!r} is not valid")

    def test_set_oauth_uid_wrong_token(self):
        self.assertFalse(self.user_id.oauth_uid)
        data = self.call_set_oauth_uid(token='fantasy')
        self.assertIn('error', data)
        self.assertEqual(data['error']['data']['message'].strip(), "Signup token 'fantasy' is not valid")

    def test_set_oauth_uid_missing_saas_auth(self):
        self.assertFalse(self.user_id.oauth_uid)
        with mute_logger('odoo.addons.saas_worker.util'):  # logs a Forbidden stacktrace
            data = self.call_set_oauth_uid(auth=False)
        self.assertIn('error', data)
        self.assertEqual(data['error']['data']['name'], 'werkzeug.exceptions.Forbidden')

    def test_set_oauth_uid(self):
        db_uuid = self.env["ir.config_parameter"].get_param("database.uuid")
        self.assertFalse(self.user_id.oauth_uid)

        data = self.call_set_oauth_uid()

        if not hasattr(self.user_id, 'invalidate_recordset'):
            # < 16.0
            self.user_id.invalidate_cache()

        self.assertEqual(data.get('result'), {'db_uuid': db_uuid}, "Invalid returned data: %r" % (data,))
        self.assertEqual(self.user_id.oauth_uid, '1985')

        clear_password = str(uuid4())
        return_value = {'result': {'is_staff_restricted': False}}
        with patch('requests.post', **{'return_value.json.return_value': return_value}) as requests_post:
            self.env(user=self.user_id)['res.users']._check_credentials(clear_password, {'interactive': True})

        requests_post.assert_called_with(
            f'{OAUTH_BASE_URL}/odoo-enterprise/trial-login',
            json={'params': {
                'user_id': '1985',
                'password': clear_password,
                'reason': '',
                'action': 'login',
                'remote_uid': self.user_id.id,
                'remote_login': self.user_id.login,
                'dbuuid': db_uuid,
                'interactive': True,
            }},
            auth=ANY,
            timeout=ANY,
        )

    def test_reset_oauth_provider_id(self):
        module_auth_oauth = self.env['ir.module.module'].search([('name', '=', 'auth_oauth')])
        if module_auth_oauth.state != 'installed':
            self.skipTest("auth_oauth is not installed")

        self.user_id.write({
            'oauth_provider_id': self.env.ref('auth_oauth.provider_facebook').id,
            'oauth_uid': '42',
        })
        self.flush_all()

        self.call_set_oauth_uid()

        if not hasattr(self.user_id, 'invalidate_recordset'):
            # < 16.0
            self.user_id.invalidate_cache()

        self.assertFalse(self.user_id.oauth_provider_id)
        self.assertEqual(self.user_id.oauth_uid, '1985')
